from enum import Enum

from .typed_ast import *


class LenImpl(PolymorphicFunction):
    def type_from_args(self, args: typing.List[Type]) -> FunctionType:
        assert (
            len(args) == 1
        ), f"'len' takes only one argument, but {len(args)} were given"
        assert isinstance(
            args[0], InstanceType
        ), "Can only determine length of instances"
        return FunctionType(args, IntegerInstanceType)

    def impl_from_args(self, args: typing.List[Type]) -> plt.AST:
        arg = args[0]
        assert isinstance(arg, InstanceType), "Can only determine length of instances"
        if arg == ByteStringInstanceType:
            return OLambda(["x"], plt.LengthOfByteString(OVar("x")))
        elif isinstance(arg.typ, ListType) or isinstance(arg.typ, DictType):
            # simple list length function
            return OLambda(
                ["x"],
                plt.FoldList(
                    OVar("x"),
                    OLambda(["a", "_"], plt.AddInteger(OVar("a"), plt.Integer(1))),
                    plt.Integer(0),
                ),
            )
        elif isinstance(arg.typ, TupleType):
            return OLambda(
                ["x"],
                plt.Integer(len(arg.typ.typs)),
            )
        raise NotImplementedError(f"'len' is not implemented for type {arg}")


class ReversedImpl(PolymorphicFunction):
    def type_from_args(self, args: typing.List[Type]) -> FunctionType:
        assert (
            len(args) == 1
        ), f"'reversed' takes only one argument, but {len(args)} were given"
        typ = args[0]
        assert isinstance(typ, InstanceType), "Can only reverse instances"
        assert isinstance(typ.typ, ListType), "Can only reverse instances of lists"
        # returns list of same type
        return FunctionType(args, typ)

    def impl_from_args(self, args: typing.List[Type]) -> plt.AST:
        arg = args[0]
        assert isinstance(arg, InstanceType), "Can only reverse instances"
        if isinstance(arg.typ, ListType):
            empty_l = empty_list(arg.typ.typ)
            return OLambda(
                ["xs"],
                plt.FoldList(
                    OVar("xs"),
                    OLambda(["a", "x"], plt.MkCons(OVar("x"), OVar("a"))),
                    empty_l,
                ),
            )
        raise NotImplementedError(f"'reversed' is not implemented for type {arg}")


class PrintImpl(PolymorphicFunction):
    def type_from_args(self, args: typing.List[Type]) -> FunctionType:
        assert all(
            isinstance(typ, InstanceType) for typ in args
        ), "Can only print instances"
        return FunctionType(args, NoneInstanceType)

    def impl_from_args(self, args: typing.List[Type]) -> plt.AST:
        if not args:
            return SafeOLambda([], plt.Trace(plt.Text("\n"), plt.NoneData()))
        assert all(
            isinstance(arg, InstanceType) for arg in args
        ), "Can only stringify instances"
        stringify_ops = [
            plt.Apply(arg.typ.stringify(), OVar(f"x{i}")) for i, arg in enumerate(args)
        ]
        stringify_ops_joined = sum(((x, plt.Text(" ")) for x in stringify_ops), ())[:-1]
        print = SafeOLambda(
            [f"x{i}" for i in range(len(args))],
            plt.Trace(plt.ConcatString(*stringify_ops_joined), plt.NoneData()),
        )
        return print


class PythonBuiltIn(Enum):
    all = OLambda(
        ["xs"],
        plt.FoldList(
            OVar("xs"),
            OLambda(["x", "a"], plt.And(OVar("x"), OVar("a"))),
            plt.Bool(True),
        ),
    )
    any = OLambda(
        ["xs"],
        plt.FoldList(
            OVar("xs"),
            OLambda(["x", "a"], plt.Or(OVar("x"), OVar("a"))),
            plt.Bool(False),
        ),
    )
    abs = OLambda(
        ["x"],
        plt.Ite(
            plt.LessThanInteger(OVar("x"), plt.Integer(0)),
            plt.Negate(OVar("x")),
            OVar("x"),
        ),
    )
    # maps an integer to a unicode code point and decodes it
    # reference: https://en.wikipedia.org/wiki/UTF-8#Encoding
    chr = OLambda(
        ["x"],
        plt.DecodeUtf8(
            plt.Ite(
                plt.LessThanInteger(OVar("x"), plt.Integer(0x0)),
                plt.TraceError("ValueError: chr() arg not in range(0x110000)"),
                plt.Ite(
                    plt.LessThanInteger(OVar("x"), plt.Integer(0x80)),
                    # encoding of 0x0 - 0x80
                    plt.ConsByteString(OVar("x"), plt.ByteString(b"")),
                    plt.Ite(
                        plt.LessThanInteger(OVar("x"), plt.Integer(0x800)),
                        # encoding of 0x80 - 0x800
                        plt.ConsByteString(
                            # we do bit manipulation using integer arithmetic here - nice
                            plt.AddInteger(
                                plt.Integer(0b110 << 5),
                                plt.DivideInteger(OVar("x"), plt.Integer(1 << 6)),
                            ),
                            plt.ConsByteString(
                                plt.AddInteger(
                                    plt.Integer(0b10 << 6),
                                    plt.ModInteger(OVar("x"), plt.Integer(1 << 6)),
                                ),
                                plt.ByteString(b""),
                            ),
                        ),
                        plt.Ite(
                            plt.LessThanInteger(OVar("x"), plt.Integer(0x10000)),
                            # encoding of 0x800 - 0x10000
                            plt.ConsByteString(
                                plt.AddInteger(
                                    plt.Integer(0b1110 << 4),
                                    plt.DivideInteger(OVar("x"), plt.Integer(1 << 12)),
                                ),
                                plt.ConsByteString(
                                    plt.AddInteger(
                                        plt.Integer(0b10 << 6),
                                        plt.DivideInteger(
                                            plt.ModInteger(
                                                OVar("x"), plt.Integer(1 << 12)
                                            ),
                                            plt.Integer(1 << 6),
                                        ),
                                    ),
                                    plt.ConsByteString(
                                        plt.AddInteger(
                                            plt.Integer(0b10 << 6),
                                            plt.ModInteger(
                                                OVar("x"), plt.Integer(1 << 6)
                                            ),
                                        ),
                                        plt.ByteString(b""),
                                    ),
                                ),
                            ),
                            plt.Ite(
                                plt.LessThanInteger(OVar("x"), plt.Integer(0x110000)),
                                # encoding of 0x10000 - 0x10FFF
                                plt.ConsByteString(
                                    plt.AddInteger(
                                        plt.Integer(0b11110 << 3),
                                        plt.DivideInteger(
                                            OVar("x"), plt.Integer(1 << 18)
                                        ),
                                    ),
                                    plt.ConsByteString(
                                        plt.AddInteger(
                                            plt.Integer(0b10 << 6),
                                            plt.DivideInteger(
                                                plt.ModInteger(
                                                    OVar("x"), plt.Integer(1 << 18)
                                                ),
                                                plt.Integer(1 << 12),
                                            ),
                                        ),
                                        plt.ConsByteString(
                                            plt.AddInteger(
                                                plt.Integer(0b10 << 6),
                                                plt.DivideInteger(
                                                    plt.ModInteger(
                                                        OVar("x"),
                                                        plt.Integer(1 << 12),
                                                    ),
                                                    plt.Integer(1 << 6),
                                                ),
                                            ),
                                            plt.ConsByteString(
                                                plt.AddInteger(
                                                    plt.Integer(0b10 << 6),
                                                    plt.ModInteger(
                                                        OVar("x"),
                                                        plt.Integer(1 << 6),
                                                    ),
                                                ),
                                                plt.ByteString(b""),
                                            ),
                                        ),
                                    ),
                                ),
                                plt.TraceError(
                                    "ValueError: chr() arg not in range(0x110000)"
                                ),
                            ),
                        ),
                    ),
                ),
            )
        ),
    )
    breakpoint = OLambda(["_"], plt.NoneData())
    hex = OLambda(
        ["x"],
        plt.DecodeUtf8(
            OLet(
                [
                    (
                        "hexlist",
                        plt.RecFun(
                            OLambda(
                                ["f", "i"],
                                plt.Ite(
                                    plt.LessThanEqualsInteger(
                                        OVar("i"), plt.Integer(0)
                                    ),
                                    plt.EmptyIntegerList(),
                                    plt.MkCons(
                                        OLet(
                                            [
                                                (
                                                    "mod",
                                                    plt.ModInteger(
                                                        OVar("i"), plt.Integer(16)
                                                    ),
                                                ),
                                            ],
                                            plt.AddInteger(
                                                OVar("mod"),
                                                plt.IfThenElse(
                                                    plt.LessThanInteger(
                                                        OVar("mod"), plt.Integer(10)
                                                    ),
                                                    plt.Integer(ord("0")),
                                                    plt.Integer(ord("a") - 10),
                                                ),
                                            ),
                                        ),
                                        plt.Apply(
                                            OVar("f"),
                                            OVar("f"),
                                            plt.DivideInteger(
                                                OVar("i"), plt.Integer(16)
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                    (
                        "mkstr",
                        OLambda(
                            ["i"],
                            plt.FoldList(
                                plt.Apply(OVar("hexlist"), OVar("i")),
                                OLambda(
                                    ["b", "i"],
                                    plt.ConsByteString(OVar("i"), OVar("b")),
                                ),
                                plt.ByteString(b""),
                            ),
                        ),
                    ),
                ],
                plt.Ite(
                    plt.EqualsInteger(OVar("x"), plt.Integer(0)),
                    plt.ByteString(b"0x0"),
                    plt.Ite(
                        plt.LessThanInteger(OVar("x"), plt.Integer(0)),
                        plt.ConsByteString(
                            plt.Integer(ord("-")),
                            plt.AppendByteString(
                                plt.ByteString(b"0x"),
                                plt.Apply(OVar("mkstr"), plt.Negate(OVar("x"))),
                            ),
                        ),
                        plt.AppendByteString(
                            plt.ByteString(b"0x"),
                            plt.Apply(OVar("mkstr"), OVar("x")),
                        ),
                    ),
                ),
            )
        ),
    )
    len = "len"
    max = OLambda(
        ["xs"],
        plt.IteNullList(
            OVar("xs"),
            plt.TraceError("ValueError: max() arg is an empty sequence"),
            plt.FoldList(
                plt.TailList(OVar("xs")),
                OLambda(
                    ["x", "a"],
                    plt.IfThenElse(
                        plt.LessThanInteger(OVar("a"), OVar("x")),
                        OVar("x"),
                        OVar("a"),
                    ),
                ),
                plt.HeadList(OVar("xs")),
            ),
        ),
    )
    min = OLambda(
        ["xs"],
        plt.IteNullList(
            OVar("xs"),
            plt.TraceError("ValueError: min() arg is an empty sequence"),
            plt.FoldList(
                plt.TailList(OVar("xs")),
                OLambda(
                    ["x", "a"],
                    plt.IfThenElse(
                        plt.LessThanInteger(OVar("a"), OVar("x")),
                        OVar("a"),
                        OVar("x"),
                    ),
                ),
                plt.HeadList(OVar("xs")),
            ),
        ),
    )
    print = "print"
    # NOTE: only correctly defined for positive y
    pow = OLambda(
        ["x", "y"],
        plt.Ite(
            plt.LessThanInteger(OVar("y"), plt.Integer(0)),
            plt.TraceError("Negative exponentiation is not supported"),
            PowImpl(OVar("x"), OVar("y")),
        ),
    )
    oct = OLambda(
        ["x"],
        plt.DecodeUtf8(
            OLet(
                [
                    (
                        "octlist",
                        plt.RecFun(
                            OLambda(
                                ["f", "i"],
                                plt.Ite(
                                    plt.LessThanEqualsInteger(
                                        OVar("i"), plt.Integer(0)
                                    ),
                                    plt.EmptyIntegerList(),
                                    plt.MkCons(
                                        plt.AddInteger(
                                            plt.ModInteger(OVar("i"), plt.Integer(8)),
                                            plt.Integer(ord("0")),
                                        ),
                                        plt.Apply(
                                            OVar("f"),
                                            OVar("f"),
                                            plt.DivideInteger(
                                                OVar("i"), plt.Integer(8)
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                    (
                        "mkoct",
                        OLambda(
                            ["i"],
                            plt.FoldList(
                                plt.Apply(OVar("octlist"), OVar("i")),
                                OLambda(
                                    ["b", "i"],
                                    plt.ConsByteString(OVar("i"), OVar("b")),
                                ),
                                plt.ByteString(b""),
                            ),
                        ),
                    ),
                ],
                plt.Ite(
                    plt.EqualsInteger(OVar("x"), plt.Integer(0)),
                    plt.ByteString(b"0o0"),
                    plt.Ite(
                        plt.LessThanInteger(OVar("x"), plt.Integer(0)),
                        plt.ConsByteString(
                            plt.Integer(ord("-")),
                            plt.AppendByteString(
                                plt.ByteString(b"0o"),
                                plt.Apply(OVar("mkoct"), plt.Negate(OVar("x"))),
                            ),
                        ),
                        plt.AppendByteString(
                            plt.ByteString(b"0o"),
                            plt.Apply(OVar("mkoct"), OVar("x")),
                        ),
                    ),
                ),
            )
        ),
    )
    range = OLambda(
        ["limit"],
        plt.Range(OVar("limit")),
    )
    reversed = "reversed"
    sum = OLambda(
        ["xs"],
        plt.FoldList(
            OVar("xs"), plt.BuiltIn(uplc.BuiltInFun.AddInteger), plt.Integer(0)
        ),
    )


PythonBuiltInTypes = {
    PythonBuiltIn.all: InstanceType(
        FunctionType(
            [InstanceType(ListType(BoolInstanceType))],
            BoolInstanceType,
        )
    ),
    PythonBuiltIn.any: InstanceType(
        FunctionType(
            [InstanceType(ListType(BoolInstanceType))],
            BoolInstanceType,
        )
    ),
    PythonBuiltIn.abs: InstanceType(
        FunctionType(
            [IntegerInstanceType],
            IntegerInstanceType,
        )
    ),
    PythonBuiltIn.chr: InstanceType(
        FunctionType(
            [IntegerInstanceType],
            StringInstanceType,
        )
    ),
    PythonBuiltIn.breakpoint: InstanceType(FunctionType([], NoneInstanceType)),
    PythonBuiltIn.len: InstanceType(PolymorphicFunctionType(LenImpl())),
    PythonBuiltIn.hex: InstanceType(
        FunctionType(
            [IntegerInstanceType],
            StringInstanceType,
        )
    ),
    PythonBuiltIn.max: InstanceType(
        FunctionType(
            [InstanceType(ListType(IntegerInstanceType))],
            IntegerInstanceType,
        )
    ),
    PythonBuiltIn.min: InstanceType(
        FunctionType(
            [InstanceType(ListType(IntegerInstanceType))],
            IntegerInstanceType,
        )
    ),
    PythonBuiltIn.print: InstanceType(PolymorphicFunctionType(PrintImpl())),
    PythonBuiltIn.pow: InstanceType(
        FunctionType(
            [IntegerInstanceType, IntegerInstanceType],
            IntegerInstanceType,
        )
    ),
    PythonBuiltIn.oct: InstanceType(
        FunctionType(
            [IntegerInstanceType],
            StringInstanceType,
        )
    ),
    PythonBuiltIn.range: InstanceType(
        FunctionType(
            [IntegerInstanceType],
            InstanceType(ListType(IntegerInstanceType)),
        )
    ),
    PythonBuiltIn.reversed: InstanceType(PolymorphicFunctionType(ReversedImpl())),
    PythonBuiltIn.sum: InstanceType(
        FunctionType(
            [InstanceType(ListType(IntegerInstanceType))],
            IntegerInstanceType,
        )
    ),
}



import hypothesis
import unittest

from hypothesis import example, given
from hypothesis import strategies as st

from . import PLUTUS_VM_PROFILE
from .utils import eval_uplc, eval_uplc_value, Unit

hypothesis.settings.load_profile(PLUTUS_VM_PROFILE)


class BuiltinTest(unittest.TestCase):
    @given(xs=st.lists(st.booleans()))
    def test_all(self, xs):
        source_code = """
def validator(x: List[bool]) -> bool:
    return all(x)
            """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(bool(ret), all(xs), "all returned wrong value")

    @given(xs=st.lists(st.booleans()))
    def test_any(self, xs):
        source_code = """
def validator(x: List[bool]) -> bool:
    return any(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(bool(ret), any(xs), "any returned wrong value")

    @given(i=st.integers())
    def test_abs(self, i):
        source_code = """
def validator(x: int) -> int:
    return abs(x)
        """
        ret = eval_uplc_value(source_code, i)
        self.assertEqual(ret, abs(i), "abs returned wrong value")

    @given(
        xs=st.one_of(
            st.lists(st.integers()), st.lists(st.integers(min_value=0, max_value=255))
        )
    )
    def test_bytes_int_list(self, xs):
        source_code = """
def validator(x: List[int]) -> bytes:
    return bytes(x)
        """
        try:
            exp = bytes(xs)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs)
        except:
            ret = None
        self.assertEqual(ret, exp, "bytes (integer list) returned wrong value")

    @given(x=st.integers(min_value=-1000, max_value=1000))
    def test_bytes_int(self, x):
        source_code = """
def validator(x: int) -> bytes:
    return bytes(x)
        """
        try:
            exp = bytes(x)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x)
        except:
            ret = None
        self.assertEqual(ret, exp, "bytes (integer) returned wrong value")

    @given(x=st.binary())
    def test_bytes_bytes(self, x):
        source_code = """
def validator(x: bytes) -> bytes:
    return bytes(x)
        """
        try:
            exp = bytes(x)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x)
        except:
            ret = None
        self.assertEqual(ret, exp, "bytes (bytes) returned wrong value")

    @given(i=st.integers())
    @example(256)
    @example(0)
    def test_chr(self, i):
        source_code = """
def validator(x: int) -> str:
    return chr(x)
        """
        # UPLC lambdas may only take one argument at a time, so we evaluate by repeatedly applying
        try:
            i_unicode = chr(i).encode("utf8")
        except (ValueError, OverflowError):
            i_unicode = None
        try:
            ret = eval_uplc_value(source_code, i)
        except:
            ret = None
        self.assertEqual(ret, i_unicode, "chr returned wrong value")

    @given(x=st.integers())
    @example(0)
    @example(-1)
    @example(100)
    def test_hex(self, x):
        source_code = """
def validator(x: int) -> str:
    return hex(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(ret.decode("utf8"), hex(x), "hex returned wrong value")

    @given(
        xs=st.one_of(
            st.builds(lambda x: str(x), st.integers()),
            st.from_regex(r"\A(?!\s).*(?<!\s)\Z", fullmatch=True),
        )
    )
    @example("")
    @example("10_00")
    @example("_")
    @example("_1")
    @example("-")
    @example("+123")
    @example("-_")
    @example("0_")
    # @example("0\n")  # stripping is broken
    def test_int_string(self, xs: str):
        source_code = """
def validator(x: str) -> int:
    return int(x)
        """
        try:
            exp = int(xs)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs.encode("utf8"))
        except Exception as e:
            ret = None
        self.assertEqual(ret, exp, "int (str) returned wrong value")

    @given(xs=st.booleans())
    def test_int_bool(self, xs: bool):
        source_code = """
def validator(x: bool) -> int:
    return int(x)
        """
        try:
            exp = int(xs)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs)
        except:
            ret = None
        self.assertEqual(ret, exp, "int (bool) returned wrong value")

    @given(xs=st.integers())
    def test_int_int(self, xs: int):
        source_code = """
def validator(x: int) -> int:
    return int(x)
        """
        try:
            exp = int(xs)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs)
        except:
            ret = None
        self.assertEqual(ret, exp, "int (int) returned wrong value")

    @given(i=st.binary())
    def test_len_bytestring(self, i):
        source_code = """
def validator(x: bytes) -> int:
    return len(x)
        """
        # UPLC lambdas may only take one argument at a time, so we evaluate by repeatedly applying
        ret = eval_uplc_value(source_code, i)
        self.assertEqual(ret, len(i), "len (bytestring) returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_len_lists(self, xs):
        source_code = """
def validator(x: List[int]) -> int:
    return len(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(ret, len(xs), "len returned wrong value")

    @given(xs=st.dictionaries(st.integers(), st.integers()))
    def test_len_dicts(self, xs):
        source_code = """
def validator(x: Dict[int, int]) -> int:
    return len(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(ret, len(xs), "len returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_len_tuples(self, xs):
        source_code = f"""
def validator(x: None) -> int:
    return len(({", ".join(map(str, xs))}{"," if len(xs) == 1 else ""}))
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(ret, len(xs), "len returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_max(self, xs):
        source_code = """
def validator(x: List[int]) -> int:
    return max(x)
        """
        try:
            exp = max(xs)
        except:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs)
        except:
            ret = None
        self.assertEqual(ret, exp, "max returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_min(self, xs):
        source_code = """
def validator(x: List[int]) -> int:
    return min(x)
        """
        try:
            exp = min(xs)
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs)
        except:
            ret = None
        self.assertEqual(ret, exp, "min returned wrong value")

    @given(x=st.integers(), y=st.integers(min_value=0, max_value=20))
    def test_pow(self, x: int, y: int):
        source_code = """
def validator(x: int, y: int) -> int:
    return pow(x, y) % 10000000000
        """
        try:
            exp = pow(x, y) % 10000000000
        except ValueError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except:
            ret = None
        self.assertEqual(ret, exp, "pow returned wrong value")

    @given(x=st.integers(), y=st.integers(min_value=-20, max_value=-1))
    def test_neg_pow(self, x: int, y: int):
        source_code = """
def validator(x: int, y: int) -> int:
    return pow(x, y) % 10000000000
        """
        try:
            eval_uplc_value(source_code, x, y)
            fail = True
        except Exception:
            fail = False
        self.assertFalse(fail, "pow worked on negative exponent")

    @given(x=st.integers())
    @example(0)
    @example(-1)
    @example(100)
    def test_oct(self, x):
        source_code = """
def validator(x: int) -> str:
    return oct(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(ret.decode("utf8"), oct(x), "oct returned wrong value")

    @given(i=st.integers(max_value=100))
    def test_range(self, i):
        source_code = """
def validator(x: int) -> List[int]:
    return range(x)
        """
        ret = eval_uplc_value(source_code, i)
        self.assertEqual(
            [x.value for x in ret], list(range(i)), "sum returned wrong value"
        )

    @given(x=st.integers())
    @example(0)
    @example(-1)
    @example(100)
    def test_str_int(self, x):
        source_code = """
def validator(x: int) -> str:
    return str(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(ret.decode("utf8"), str(x), "str returned wrong value")

    @given(x=st.booleans())
    def test_str_bool(self, x):
        source_code = """
def validator(x: bool) -> str:
    return str(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(ret.decode("utf8"), str(x), "str returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_sum(self, xs):
        source_code = """
def validator(x: List[int]) -> int:
    return sum(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(ret, sum(xs), "sum returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_reversed(self, xs):
        source_code = """
def validator(x: List[int]) -> List[int]:
    return reversed(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(
            [r.value for r in ret], list(reversed(xs)), "reversed returned wrong value"
        )

    @given(x=st.integers())
    def test_bool_constr_int(self, x):
        source_code = """
def validator(x: int) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), bool(x), "bool (int) returned wrong value")

    @given(x=st.text())
    def test_bool_constr_str(self, x):
        source_code = """
def validator(x: str) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, x.encode("utf8"))
        self.assertEqual(bool(ret), bool(x), "bool (str) returned wrong value")

    @given(x=st.binary())
    def test_bool_constr_bytes(self, x):
        source_code = """
def validator(x: bytes) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), bool(x), "bool (bytes) returned wrong value")

    @given(x=st.none())
    def test_bool_constr_none(self, x):
        source_code = """
def validator(x: None) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(bool(ret), bool(x), "bool (none) returned wrong value")

    @given(x=st.booleans())
    def test_bool_constr_bool(self, x):
        source_code = """
def validator(x: bool) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), bool(x), "bool (bool) returned wrong value")

    @given(xs=st.lists(st.integers()))
    def test_bool_constr_list(self, xs):
        source_code = """
def validator(x: List[int]) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(bool(ret), bool(xs), "bool (list) returned wrong value")

    @given(xs=st.dictionaries(st.integers(), st.binary()))
    def test_bool_constr_dict(self, xs):
        source_code = """
def validator(x: Dict[int, str]) -> bool:
    return bool(x)
        """
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(bool(ret), bool(xs), "bool (list) returned wrong value")

    @given(x=st.integers(), y=st.booleans(), z=st.none())
    def test_print_compile(self, x, y, z):
        source_code = """
def validator(x: int, y: bool, z: None) -> None:
    print(x, y, z)
        """
        eval_uplc(source_code, x, y, Unit())



""" An implementation of some math operations in opshin """


def gcd(a: int, b: int) -> int:
    while b != 0:
        a, b = b, a % b
    return abs(a)


def sign(a: int) -> int:
    return -1 if a < 0 else 1


def unsigned_int_from_bytes_big(b: bytes) -> int:
    """Converts a bytestring into the corresponding integer, big/network byteorder, unsigned"""
    acc = 0
    for i in range(len(b)):
        acc = acc * 256 + b[i]
    return acc


def bytes_big_from_unsigned_int(b: int) -> bytes:
    """Converts an integer into the corresponding bytestring, big/network byteorder, unsigned"""
    assert b >= 0
    if b == 0:
        return b"\x00"
    acc = b""
    while b > 0:
        acc = bytes([b % 256]) + acc
        b //= 256
    return acc


def ceil(a: int, b: int):
    """Returns a divided by b rounded towards positive infinity"""
    return (a + b - 1) // b if b > 0 else (a + b + 1) // b


def floor(a: int, b: int):
    """Returns a divided by b rounded towards negative infinity"""
    return a // b



import fractions

import hypothesis
import hypothesis.strategies as hst

from opshin.std import math as oc_math

import math

pos_int = hst.integers(min_value=0)


@hypothesis.given(pos_int, pos_int)
def test_gcd(a: int, b: int):
    assert oc_math.gcd(a, b) == math.gcd(a, b), "Invalid gcd"


@hypothesis.given(hst.integers())
def test_sign(a: int):
    assert oc_math.sign(a) == math.copysign(1, a), "Invalid sign"


@hypothesis.given(hst.binary())
def test_unsigned_int_from_bytes_big(b: bytes):
    assert oc_math.unsigned_int_from_bytes_big(b) == int.from_bytes(
        b, byteorder="big", signed=False
    ), "Invalid from bytes"


@hypothesis.given(hst.integers())
@hypothesis.example(1000)
def test_bytes_big_from_unsigned_int(b: int):
    try:
        res = oc_math.bytes_big_from_unsigned_int(b)
    except AssertionError:
        res = None
    try:
        exp = b.to_bytes(
            max(1, (b.bit_length() + 7) // 8), byteorder="big", signed=False
        )
    except OverflowError:
        exp = None
    assert res == exp, "Invalid to bytes"


@hypothesis.given(hst.integers(), hst.integers())
def test_ceil(a: int, b: int):
    hypothesis.assume(b != 0)
    assert oc_math.ceil(a, b) == math.ceil(fractions.Fraction(a, b)), "Invalid ceil"


@hypothesis.given(hst.integers(), hst.integers())
def test_floor(a: int, b: int):
    hypothesis.assume(b != 0)
    assert oc_math.floor(a, b) == math.floor(fractions.Fraction(a, b)), "Invalid floor"



from parameterized import parameterized

from uplc import ast as uplc, eval as uplc_eval
from ..utils import eval_uplc


@parameterized.expand(
    [
        [[0, 1]],
        [[0]],
        [[0, 1, 2]],
        [[b"hello", 0]],
    ]
)
def test_integrity_check(xs):
    source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from opshin.std.integrity import check_integrity

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: B) -> None:
    check_integrity(x)
"""
    obj = uplc.PlutusConstr(
        1,
        [
            uplc.PlutusInteger(x) if isinstance(x, int) else uplc.PlutusByteString(x)
            for x in xs
        ],
    )
    try:
        eval_uplc(source_code, obj)
    except Exception as e:
        print(e)
        res = False
    else:
        res = True
    assert res == (len(xs) == 2 and all(isinstance(x, int) for x in xs))


@parameterized.expand(
    [
        [0],
        [1],
        [2],
    ]
)
def test_integrity_check_list(bar_constr):
    source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from opshin.std.integrity import check_integrity

@dataclass()
class A(PlutusData):
    CONSTR_ID = 1
    
@dataclass()
class C(PlutusData):
    CONSTR_ID = 0

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    bar: Union[A, C]

def validator(x: B) -> None:
    check_integrity(x)
"""
    obj = uplc.PlutusConstr(
        1,
        [uplc.PlutusConstr(bar_constr, [])],
    )
    try:
        eval_uplc(source_code, obj)
    except:
        res = False
    else:
        res = True
    assert res == (bar_constr in (0, 1))


@parameterized.expand(
    [
        [[0, 1], [1, 1, 1]],
        [[b"hello"], [1, 1]],
        [[0, 1, 2], [1, 0]],
    ]
)
def test_integrity_check_list(foobar, bar_constrs):
    source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from opshin.std.integrity import check_integrity

@dataclass()
class A(PlutusData):
    CONSTR_ID = 1

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: List[int]
    bar: List[A]

def validator(x: B) -> None:
    check_integrity(x)
"""
    obj = uplc.PlutusConstr(
        1,
        [
            uplc.PlutusList(
                [
                    (
                        uplc.PlutusInteger(x)
                        if isinstance(x, int)
                        else uplc.PlutusByteString(x)
                    )
                    for x in foobar
                ]
            ),
            uplc.PlutusList([uplc.PlutusConstr(c, []) for c in bar_constrs]),
        ],
    )
    try:
        eval_uplc(source_code, obj)
    except:
        res = False
    else:
        res = True
    assert res == (
        all(isinstance(x, int) for x in foobar) and all(c == 1 for c in bar_constrs)
    )


@parameterized.expand(
    [
        [[0, 1, 2], [1, 1, 1]],
        # check for incorrect type in keys
        [[b"hello", 1], [1, 1]],
        # check for incorrect type in values
        [[1, 2], [1, b"hello"]],
    ]
)
def test_integrity_check_dict(keys, values):
    source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from opshin.std.integrity import check_integrity

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: Dict[int, int]

def validator(x: B) -> None:
    check_integrity(x)
"""
    obj = uplc.PlutusConstr(
        1,
        [
            uplc.PlutusMap(
                {
                    (
                        uplc.PlutusInteger(x)
                        if isinstance(x, int)
                        else uplc.PlutusByteString(x)
                    ): (
                        uplc.PlutusInteger(y)
                        if isinstance(y, int)
                        else uplc.PlutusByteString(y)
                    )
                    for x, y in zip(keys, values)
                },
            ),
        ],
    )
    try:
        eval_uplc(source_code, obj)
    except:
        res = False
    else:
        res = True
    assert res == (
        all(isinstance(x, int) for x in keys + values) and len(set(keys)) == len(keys)
    )


def test_integrity_check_rename():
    source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from opshin.std.integrity import check_integrity as integ

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int

def validator(x: B) -> None:
    integ(x)
"""
    obj = uplc.PlutusConstr(
        1,
        [uplc.PlutusInteger(1)],
    )
    eval_uplc(source_code, obj)


# TODO implement better way to check for uniqueness test in dict keys



import unittest

import cbor2
import hypothesis
import pycardano
from frozendict import frozendict
import frozenlist as fl
from hypothesis import example, given
from hypothesis import strategies as st

from uplc import ast as uplc, eval as uplc_eval
from uplc.ast import (
    PlutusMap,
    PlutusConstr,
    PlutusList,
    PlutusInteger,
    PlutusByteString,
)

from . import PLUTUS_VM_PROFILE
from .utils import eval_uplc, eval_uplc_value, Unit
from .. import compiler

hypothesis.settings.load_profile(PLUTUS_VM_PROFILE)

from opshin.ledger.api_v2 import (
    FinitePOSIXTime,
    PosInfPOSIXTime,
    UpperBoundPOSIXTime,
    FalseData,
    TrueData,
)


def frozenlist(l):
    l = fl.FrozenList(l)
    l.freeze()
    return l


pos_int = st.integers(min_value=0, max_value=2**64 - 1)


uplc_data_integer = st.builds(PlutusInteger, st.integers())
uplc_data_bytestring = st.builds(PlutusByteString, st.binary())


def rec_data_strategies(uplc_data):
    uplc_data_list = st.builds(lambda x: PlutusList(frozenlist(x)), st.lists(uplc_data))
    uplc_data_constr = st.builds(
        lambda x, y: PlutusConstr(x, frozenlist(y)),
        pos_int,
        st.lists(uplc_data),
    )
    uplc_data_map = st.builds(
        PlutusMap,
        st.dictionaries(
            st.one_of(
                uplc_data_integer, uplc_data_bytestring
            ),  # TODO technically constr is legal too, but causes hashing error
            uplc_data,
            dict_class=frozendict,
        ),
    )
    return st.one_of(uplc_data_map, uplc_data_list, uplc_data_constr)


uplc_data = st.recursive(
    st.one_of(uplc_data_bytestring, uplc_data_integer),
    rec_data_strategies,
    max_leaves=4,
)
uplc_data_list = st.builds(lambda x: PlutusList(frozenlist(x)), st.lists(uplc_data))

# TODO fix handling of these strings
formattable_text = st.from_regex(r"\A((?!['\\])[ -~])*\Z")


class OpTest(unittest.TestCase):
    @given(x=st.booleans(), y=st.booleans())
    def test_and_bool(self, x, y):
        source_code = """
def validator(x: bool, y: bool) -> bool:
    return x and y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(bool(ret), x and y, "and returned wrong value")

    @given(x=st.booleans(), y=st.booleans())
    def test_or_bool(self, x, y):
        source_code = """
def validator(x: bool, y: bool) -> bool:
    return x or y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(bool(ret), x or y, "or returned wrong value")

    @given(x=st.booleans())
    def test_not_bool(self, x):
        source_code = """
def validator(x: bool) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), not x, "not returned wrong value")

    @given(x=st.integers())
    def test_usub_int(self, x):
        source_code = """
def validator(x: int) -> int:
    return -x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(ret, -x, "not returned wrong value")

    @given(x=st.integers())
    def test_uadd_int(self, x):
        source_code = """
def validator(x: int) -> int:
    return +x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(ret, +x, "not returned wrong value")

    @given(x=st.integers())
    def test_not_int(self, x):
        source_code = """
def validator(x: int) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), not x, "not returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_add_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x + y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x + y, "+ returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_sub_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x - y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x - y, "- returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_mul_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x * y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x * y, "* returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_div_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x // y
            """
        try:
            exp = x // y
        except ZeroDivisionError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except Exception:
            ret = None
        self.assertEqual(ret, exp, "// returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_mod_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x % y
            """
        try:
            exp = x % y
        except ZeroDivisionError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except Exception:
            ret = None
        self.assertEqual(ret, exp, "% returned wrong value")

    @given(x=st.integers(), y=st.integers(min_value=0, max_value=20))
    def test_pow_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x ** y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x**y, "** returned wrong value")

    @given(x=st.integers(), y=st.integers(min_value=-20, max_value=-1))
    def test_neg_pow_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> int:
    return x ** y
            """
        try:
            eval_uplc_value(source_code, x, y)
            fail = True
        except Exception:
            fail = False
        self.assertFalse(fail, "** worked with negative exponent")

    @given(x=st.binary(), y=st.binary())
    def test_add_bytes(self, x, y):
        source_code = """
def validator(x: bytes, y: bytes) -> bytes:
    return x + y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x + y, "+ returned wrong value")

    @given(x=st.text(), y=st.text())
    def test_add_str(self, x, y):
        source_code = """
def validator(x: str, y: str) -> str:
    return x + y
            """
        ret = eval_uplc_value(source_code, x.encode("utf8"), y.encode("utf8")).decode(
            "utf8"
        )
        self.assertEqual(ret, x + y, "+ returned wrong value")

    @given(x=st.binary(), y=st.integers(), z=st.integers())
    @example(b"\x00", -2, 0)
    @example(b"1234", 1, 2)
    @example(b"1234", 2, 4)
    @example(b"1234", 2, 2)
    @example(b"1234", 3, 3)
    @example(b"1234", 3, 1)
    def test_slice_bytes(self, x, y, z):
        source_code = """
def validator(x: bytes, y: int, z: int) -> bytes:
    return x[y:z]
            """
        try:
            exp = x[y:z]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y, z)
        except:
            ret = None
        self.assertEqual(ret, exp, "byte slice returned wrong value")

    @given(x=st.binary(), y=st.integers())
    @example(b"\x00", -2)
    @example(b"1234", 1)
    @example(b"1234", 2)
    @example(b"1234", 2)
    @example(b"1234", 3)
    @example(b"1234", 3)
    def test_slice_bytes_lower(self, x, y):
        source_code = """
def validator(x: bytes, y: int) -> bytes:
    return x[y:]
            """
        try:
            exp = x[y:]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except:
            ret = None
        self.assertEqual(ret, exp, "byte slice returned wrong value")

    @given(x=st.binary(), y=st.integers())
    @example(b"\x00", 0)
    @example(b"1234", 2)
    @example(b"1234", 4)
    @example(b"1234", 2)
    @example(b"1234", 3)
    @example(b"1234", 1)
    def test_slice_bytes_upper(self, x, y):
        source_code = """
def validator(x: bytes, y: int) -> bytes:
    return x[:y]
            """
        try:
            exp = x[:y]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except:
            ret = None
        self.assertEqual(ret, exp, "byte slice returned wrong value")

    @given(x=st.binary())
    @example(b"\x00")
    @example(b"1234")
    @example(b"1234")
    @example(b"1234")
    @example(b"1234")
    @example(b"1234")
    def test_slice_bytes_full(self, x):
        source_code = """
def validator(x: bytes) -> bytes:
    return x[:]
            """
        try:
            exp = x[:]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x)
        except:
            ret = None
        self.assertEqual(ret, exp, "byte slice returned wrong value")

    @given(x=st.binary(), y=st.integers())
    @example(b"1234", 0)
    @example(b"1234", 1)
    @example(b"1234", -1)
    def test_index_bytes(self, x, y):
        source_code = """
def validator(x: bytes, y: int) -> int:
    return x[y]
            """
        try:
            exp = x[y]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except:
            ret = None
        self.assertEqual(ret, exp, "byte index returned wrong value")

    @given(xs=st.lists(st.integers()), y=st.integers())
    @example(xs=[0], y=-1)
    @example(xs=[0], y=0)
    def test_index_list(self, xs, y):
        source_code = """
from typing import Dict, List, Union
def validator(x: List[int], y: int) -> int:
    return x[y]
            """
        try:
            exp = xs[y]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs, y)
        except Exception as e:
            ret = None
        self.assertEqual(ret, exp, "list index returned wrong value")

    @given(x=st.lists(st.integers(), max_size=20), y=st.integers(), z=st.integers())
    @example([0], -2, 0)
    @example([1, 2, 3, 4], 1, 2)
    @example([1, 2, 3, 4], 2, 4)
    @example([1, 2, 3, 4], 2, 2)
    @example([1, 2, 3, 4], 3, 3)
    @example([1, 2, 3, 4], 3, 1)
    def test_slice_list(self, x, y, z):
        source_code = """
def validator(x: List[int], y: int, z: int) -> List[int]:
    return x[y:z]
            """
        try:
            exp = x[y:z]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y, z)
        except:
            ret = None
        self.assertEqual(
            ret,
            [PlutusInteger(x) for x in exp] if exp is not None else exp,
            "list slice returned wrong value",
        )

    @given(x=st.lists(st.integers(), max_size=20), y=st.integers())
    @example([0], -2)
    @example([1, 2, 3, 4], 1)
    @example([1, 2, 3, 4], 2)
    @example([1, 2, 3, 4], 2)
    @example([1, 2, 3, 4], 3)
    @example([1, 2, 3, 4], 3)
    def test_slice_list_lower(self, x, y):
        source_code = """
def validator(x: List[int], y: int) -> List[int]:
    return x[y:]
            """
        try:
            exp = x[y:]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except:
            ret = None
        self.assertEqual(
            ret,
            [PlutusInteger(x) for x in exp] if exp is not None else exp,
            "list slice returned wrong value",
        )

    @given(x=st.lists(st.integers(), max_size=20), y=st.integers())
    @example([0], 0)
    @example([1, 2, 3, 4], 2)
    @example([1, 2, 3, 4], 4)
    @example([1, 2, 3, 4], 2)
    @example([1, 2, 3, 4], 3)
    @example([1, 2, 3, 4], 1)
    def test_slice_list_upper(self, x, y):
        source_code = """
def validator(x: List[int], y: int) -> List[int]:
    return x[:y]
            """
        try:
            exp = x[:y]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x, y)
        except:
            ret = None
        self.assertEqual(
            ret,
            [PlutusInteger(x) for x in exp] if exp is not None else exp,
            "list slice returned wrong value",
        )

    @given(x=st.lists(st.integers(), max_size=20))
    @example([0])
    @example([1, 2, 3, 4])
    @example([1, 2, 3, 4])
    @example([1, 2, 3, 4])
    @example([1, 2, 3, 4])
    @example([1, 2, 3, 4])
    def test_slice_list_full(self, x):
        source_code = """
def validator(x: List[int]) -> List[int]:
    return x[:]
            """
        try:
            exp = x[:]
        except IndexError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, x)
        except:
            ret = None
        self.assertEqual(
            ret,
            [PlutusInteger(x) for x in exp] if exp is not None else exp,
            "list slice returned wrong value",
        )

    @given(xs=st.lists(st.integers()), y=st.integers())
    @example(xs=[0, 1], y=-1)
    @example(xs=[0, 1], y=0)
    def test_in_list_int(self, xs, y):
        source_code = """
from typing import Dict, List, Union
def validator(x: List[int], y: int) -> bool:
    return y in x
            """
        ret = eval_uplc_value(source_code, xs, y)
        self.assertEqual(ret, y in xs, "list in returned wrong value")

    @given(xs=st.lists(st.binary()), y=st.binary())
    def test_in_list_bytes(self, xs, y):
        source_code = """
from typing import Dict, List, Union
def validator(x: List[bytes], y: bytes) -> bool:
    return y in x
            """
        ret = eval_uplc_value(source_code, xs, y)
        self.assertEqual(ret, y in xs, "list in returned wrong value")

    @given(xs=uplc_data_list, y=uplc_data)
    def test_in_list_data(self, xs, y):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
def validator(x: List[Anything], y: Anything) -> bool:
    return y in x
            """
        ret = eval_uplc_value(source_code, xs, y)
        self.assertEqual(ret, y in xs.value, "list in returned wrong value")

    @given(xs=uplc_data_list, y=uplc_data)
    def test_not_in_list_data(self, xs, y):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
def validator(x: List[Anything], y: Anything) -> bool:
    return y not in x
            """
        ret = eval_uplc_value(source_code, xs, y)
        self.assertEqual(ret, y not in xs.value, "list not in returned wrong value")

    @given(xs=st.lists(st.integers()), y=st.integers())
    @example(xs=[0, 1], y=-1)
    @example(xs=[0, 1], y=0)
    def test_not_in_list_int(self, xs, y):
        source_code = """
from typing import Dict, List, Union
def validator(x: List[int], y: int) -> bool:
    return y not in x
            """
        ret = eval_uplc_value(source_code, xs, y)
        self.assertEqual(ret, y not in xs, "list not in returned wrong value")

    @given(xs=st.lists(st.binary()), y=st.binary())
    def test_not_in_list_bytes(self, xs, y):
        source_code = """
from typing import Dict, List, Union
def validator(x: List[bytes], y: bytes) -> bool:
    return y not in x
            """
        ret = eval_uplc_value(source_code, xs, y)
        self.assertEqual(ret, y not in xs, "list not in returned wrong value")

    @given(x=st.lists(st.integers()))
    def test_not_list(self, x):
        source_code = """
def validator(x: List[int]) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), not x, "not returned wrong value")

    @given(x=st.binary(), y=st.binary())
    def test_eq_bytes(self, x, y):
        source_code = """
def validator(x: bytes, y: bytes) -> bool:
    return x == y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x == y, "bytes eq returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_eq_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> bool:
    return x == y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x == y, "int eq returned wrong value")

    @given(x=st.text(), y=st.text())
    def test_eq_str(self, x, y):
        source_code = """
def validator(x: str, y: str) -> bool:
    return x == y
            """
        ret = eval_uplc_value(source_code, x.encode("utf8"), y.encode("utf8"))
        self.assertEqual(ret, x == y, "str eq returned wrong value")

    @given(x=st.booleans(), y=st.booleans())
    def test_eq_bool(self, x, y):
        source_code = """
def validator(x: bool, y: bool) -> bool:
    return x == y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x == y, "bool eq returned wrong value")

    @given(x=uplc_data, y=uplc_data)
    def test_eq_data(self, x, y):
        source_code = """
from opshin.prelude import *

def validator(x: Anything, y: Anything) -> bool:
    return x == y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x == y, "any eq returned wrong value")

    @given(x=st.booleans(), y=st.integers())
    def test_eq_bool_int(self, x, y):
        source_code = """
def validator(x: bool, y: int) -> bool:
    return x == y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x == y, "bool eq int returned wrong value")

    @given(x=st.text(), y=st.text())
    def test_neq_str(self, x, y):
        source_code = """
def validator(x: str, y: str) -> bool:
    return x != y
            """
        ret = eval_uplc_value(source_code, x.encode("utf8"), y.encode("utf8"))
        self.assertEqual(ret, x != y, "str neq returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_neq_int(self, x, y):
        source_code = """
def validator(x: int, y: int) -> bool:
    return x != y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x != y, "int neq returned wrong value")

    @given(x=st.binary(), y=st.binary())
    def test_neq_bytes(self, x, y):
        source_code = """
def validator(x: bytes, y: bytes) -> bool:
    return x != y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x != y, "bytes neq returned wrong value")

    @given(x=st.booleans(), y=st.booleans())
    def test_neq_bool(self, x, y):
        source_code = """
def validator(x: bool, y: bool) -> bool:
    return x != y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x != y, "bool neq returned wrong value")

    @given(x=st.booleans(), y=st.integers())
    def test_neq_bool_int(self, x, y):
        source_code = """
def validator(x: bool, y: int) -> bool:
    return x != y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x != y, "bool neq int returned wrong value")

    @given(x=uplc_data, y=uplc_data)
    def test_neq_data(self, x, y):
        source_code = """
from opshin.prelude import *

def validator(x: Anything, y: Anything) -> bool:
    return x != y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x != y, "any neq returned wrong value")

    @given(x=st.integers(min_value=0, max_value=150), y=st.text())
    def test_mul_int_str(self, x, y):
        source_code = """
def validator(x: int, y: str) -> str:
    return x * y
            """
        ret = eval_uplc_value(source_code, x, y.encode("utf8"))
        self.assertEqual(ret.decode("utf8"), x * y, "* returned wrong value")

    @given(x=st.text(), y=st.integers(min_value=0, max_value=150))
    def test_mul_str_int(self, x, y):
        source_code = """
def validator(x: str, y: int) -> str:
    return x * y
            """
        ret = eval_uplc_value(source_code, x.encode("utf8"), y)
        self.assertEqual(ret.decode("utf8"), x * y, "* returned wrong value")

    @given(x=st.integers(min_value=0, max_value=150), y=st.binary())
    def test_mul_int_bytes(self, x, y):
        source_code = """
def validator(x: int, y: bytes) -> bytes:
    return x * y
        """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x * y, "bytes int multiplication returned wrong value")

    @given(x=st.binary(), y=st.integers(min_value=0, max_value=150))
    def test_mul_bytes_int(self, x, y):
        source_code = """
def validator(x: bytes, y: int) -> bytes:
    return x * y
            """
        ret = eval_uplc_value(source_code, x, y)
        self.assertEqual(ret, x * y, "* returned wrong value")

    @given(
        xs=st.lists(st.integers(), max_size=20), ys=st.lists(st.integers(), max_size=20)
    )
    def test_add_list(self, xs, ys):
        source_code = """
def validator(x: List[int], y: List[int]) -> List[int]:
    return x + y
            """
        ret = eval_uplc_value(source_code, xs, ys)
        self.assertEqual(
            ret,
            [PlutusInteger(x) for x in xs] + [PlutusInteger(y) for y in ys],
            "+ returned wrong value",
        )

    @given(x=st.integers())
    def test_fmt_int(self, x):
        source_code = """
def validator(x: int) -> str:
    return f"{x}"
    """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(
            ret.decode("utf8"), f"{x}", "int string formatting returned wrong value"
        )

    @given(x=st.booleans())
    def test_fmt_bool(self, x):
        source_code = """
def validator(x: bool) -> str:
    return f"{x}"
    """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(
            ret.decode("utf8"), f"{x}", "bool string formatting returned wrong value"
        )

    @given(x=st.text())
    def test_fmt_str(self, x):
        source_code = """
def validator(x: str) -> str:
    return f"{x}"
            """
        ret = eval_uplc_value(source_code, x.encode("utf8"))
        self.assertEqual(
            ret.decode("utf8"), f"{x}", "string string formatting returned wrong value"
        )

    @given(x=st.binary())
    @example(b"'")
    @example(b'"')
    @example(b"\\")
    @example(b"\r")
    @example(b"\n")
    @example(b"\t")
    @example(b"\x7f")
    def test_fmt_bytes(self, x):
        source_code = """
def validator(x: bytes) -> str:
    return f"{x}"
        """
        ret = eval_uplc_value(source_code, x).decode("utf8")
        if b"'" not in x:
            self.assertEqual(
                ret, f"{x}", "bytes string formatting returned wrong value"
            )
        else:
            # NOTE: formally this is a bug where we do not have the same semantics as python
            # specifically when ' is contained in the string we do not change the quotation marks
            self.assertEqual(
                eval(ret), x, "bytes string formatting returned wrong value"
            )

    @given(x=st.none())
    def test_fmt_none(self, x):
        source_code = """
def validator(x: None) -> str:
    return f"{x}"
            """
        ret = eval_uplc_value(source_code, Unit()).decode("utf8")
        self.assertEqual(ret, f"{x}", "none string formatting returned wrong value")

    @given(
        x=st.builds(
            UpperBoundPOSIXTime,
            st.one_of(
                st.builds(FinitePOSIXTime, st.integers()), st.builds(PosInfPOSIXTime)
            ),
            st.one_of(st.builds(TrueData), st.builds(FalseData)),
        )
    )
    @example(UpperBoundPOSIXTime(PosInfPOSIXTime(), TrueData()))
    def test_fmt_dataclass(self, x: UpperBoundPOSIXTime):
        source_code = """
from opshin.prelude import *

def validator(x: UpperBoundPOSIXTime) -> str:
    return f"{x}"
            """

        ret = eval_uplc_value(source_code, x).decode("utf8")
        self.assertEqual(
            ret, f"{x}", "several element string formatting returned wrong value"
        )

    @given(x=st.integers(), y=st.integers())
    def test_fmt_multiple(self, x, y):
        source_code = """
def validator(x: int, y: int) -> str:
    return f"a{x}b{y}c"
            """
        ret = eval_uplc_value(source_code, x, y).decode("utf8")
        self.assertEqual(
            ret, f"a{x}b{y}c", "several element string formatting returned wrong value"
        )

    @given(x=st.lists(st.integers()))
    @example([])
    @example([0])
    def test_fmt_tuple_int(self, x):
        params = [f"a{i}" for i in range(len(x))]
        source_code = f"""
def validator({",".join(p + ": int" for p in params)}) -> str:
    return f"{{({"".join(p + "," for p in params)})}}"
            """
        ret = eval_uplc_value(source_code, *x if x else [0]).decode("utf8")
        exp = f"{tuple(x)}"
        self.assertEqual(
            ret, exp, "integer tuple string formatting returned wrong value"
        )

    @given(xs=st.lists(formattable_text))
    def test_fmt_tuple_str(self, xs):
        # TODO strings are not properly escaped here
        params = [f"a{i}" for i in range(len(xs))]
        source_code = f"""
def validator({",".join(p + ": str" for p in params)}) -> str:
    return f"{{({"".join(p + "," for p in params)})}}"
            """
        ret = eval_uplc_value(
            source_code, *(x.encode("utf8") for x in xs) if xs else [0]
        ).decode("utf8")
        exp = f"{tuple(xs)}"
        self.assertEqual(ret, exp, "tuple string formatting returned wrong value")

    @given(x=st.integers(), y=st.integers())
    def test_fmt_pair_int(self, x, y):
        source_code = f"""
def validator(x: int, y: int) -> str:
    a = ""
    for p in {{x:y}}.items():
        a = f"{{p}}"
    return a
            """
        exp = f"{(x, y)}"
        ret = eval_uplc_value(source_code, x, y).decode("utf8")
        self.assertEqual(
            ret, exp, "integer tuple string formatting returned wrong value"
        )

    @given(x=formattable_text, y=formattable_text)
    def test_fmt_pair_str(self, x, y):
        # TODO strings are not properly escaped here
        source_code = f"""
def validator(x: str, y: str) -> str:
    a = ""
    for p in {{x:y}}.items():
        a = f"{{p}}"
    return a
            """
        exp = f"{(x, y)}"
        ret = eval_uplc_value(source_code, x.encode("utf8"), y.encode("utf8")).decode(
            "utf8"
        )
        self.assertEqual(
            ret, exp, "string tuple string formatting returned wrong value"
        )

    @given(xs=st.lists(formattable_text))
    @example([])
    @example(["x"])
    def test_fmt_list_str(self, xs):
        # TODO strings are not properly escaped here
        source_code = """
from opshin.prelude import *

def validator(x: List[str]) -> str:
    return f"{x}"
            """
        exp = f"{list(xs)}"
        ret = eval_uplc_value(source_code, [x.encode("utf8") for x in xs]).decode(
            "utf8"
        )
        self.assertEqual(ret, exp, "string list string formatting returned wrong value")

    @given(xs=st.lists(st.integers()))
    @example([])
    @example([0])
    def test_fmt_list_int(self, xs):
        source_code = """
from opshin.prelude import *

def validator(x: List[int]) -> str:
    return f"{x}"
            """
        exp = f"{list(xs)}"
        ret = eval_uplc_value(source_code, xs).decode("utf8")
        self.assertEqual(
            ret, exp, "integer list string formatting returned wrong value"
        )

    @given(x=st.dictionaries(st.integers(), st.integers()))
    def test_not_dict(self, x):
        source_code = """
def validator(x: Dict[int, int]) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), not x, "not returned wrong value")

    @given(xs=st.dictionaries(st.integers(), st.integers()), y=st.integers())
    def test_index_dict(self, xs, y):
        source_code = """
from typing import Dict, List, Union
def validator(x: Dict[int, int], y: int) -> int:
    return x[y]
            """
        try:
            exp = xs[y]
        except KeyError:
            exp = None
        try:
            ret = eval_uplc_value(source_code, xs, y)
        except Exception as e:
            ret = None
        self.assertEqual(ret, exp, "list index returned wrong value")

    @given(xs=st.dictionaries(formattable_text, st.integers()))
    @example(dict())
    @example({"": 0})
    def test_fmt_dict_int(self, xs):
        # TODO strings are not properly escaped here
        source_code = """
from opshin.prelude import *

def validator(x: Dict[str, int]) -> str:
    return f"{x}"
            """
        exp = f"{dict(xs)}"
        ret = eval_uplc_value(
            source_code, {k.encode("utf8"): v for k, v in xs.items()}
        ).decode("utf8")
        self.assertEqual(ret, exp, "dict string formatting returned wrong value")

    @given(x=uplc_data)
    @example(PlutusByteString(b""))
    @example(PlutusConstr(0, [PlutusByteString(b"'")]))
    @example(
        PlutusMap({PlutusInteger(1): PlutusMap({}), PlutusInteger(0): PlutusMap({})})
    )
    def test_fmt_any(self, x):
        x_cbor = uplc.plutus_cbor_dumps(x)
        x_data = pycardano.RawPlutusData(cbor2.loads(x_cbor))
        source_code = """
def validator(x: Anything) -> str:
    return f"{x}"
            """
        # NOTE: this is technically a deviation from the semantics of pycardano but I expect the pycardano semantics to change soon
        exp = f"RawPlutusData(data={repr(x_data.data)})"
        ret = eval_uplc_value(source_code, pycardano.RawCBOR(x_cbor)).decode("utf8")
        if "\\'" in ret:
            RawPlutusData = pycardano.RawPlutusData
            CBORTag = cbor2.CBORTag
            self.assertEqual(
                eval(ret), x_data, "raw cbor string formatting returned wrong value"
            )
        else:
            self.assertEqual(
                ret, exp, "raw cbor string formatting returned wrong value"
            )

    @given(x=st.text())
    def test_not_string(self, x):
        source_code = """
def validator(x: str) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, x.encode("utf8"))
        self.assertEqual(bool(ret), not x, "not returned wrong value")

    @given(x=st.binary())
    def test_not_bytes(self, x):
        source_code = """
def validator(x: bytes) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, x)
        self.assertEqual(bool(ret), not x, "not returned wrong value")

    def test_not_unit(self):
        source_code = """
def validator(x: None) -> bool:
    return not x
            """
        ret = eval_uplc_value(source_code, uplc.BuiltinUnit())
        self.assertEqual(bool(ret), not None, "not returned wrong value")
